@@ -29,14 +29,14 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <ctype.h>
+#include <time.h>
 
-#include <glib.h>
-
-#include "exelf.h"
+#include "exmap.h"
 
 #define EXIT_KERNEL_ERROR  -1
 #define EXIT_LOGFILE_ERROR -2
 #define EXIT_SYSDATA_ERROR -3
+#define EXIT_INVALID_ARGS  -4
 
 #define EXMFILE "/proc/exmap"
 #define BUF_SIZE 256
@@ -56,104 +56,28 @@
 #define SCALE_U(x) "B"
 #endif
 
-#define SORT_BY_MASK   0x000000ff
-#define SORT_BY_RESET  ~SORT_BY_MASK
-#define SORT_BY_NONE   0
-#define SORT_BY_VM     1
-#define SORT_BY_SOLE   2
-#define SORT_BY_HEAP   3
-#define SORT_BY_STACK  4
-#define SORT_BY_MAPPED 5
+#ifdef INTERACTIVE
+extern int rows;
+extern int columns;
 
-#define SORT_BY_GET(x)    (x & SORT_BY_MASK)
-#define SORT_BY_SET(x, v) do{x &= SORT_BY_RESET; x |= v;}while(0)
+#define myprintf(args...)                                   \
+do {                                                        \
+    if (of == stdout && global_lines >= rows - 1)           \
+      {                                                     \
+        global_lines = 0;                                   \
+        fprintf (of, "  --- Press Enter for more ---");     \
+        fgetc (stdin);                                      \
+      }                                                     \
+                                                            \
+    fprintf (args);                                         \
+    global_lines++;                                         \
+}while (0)
 
-/* sort-type: decending or ascending */
-#define SORT_TYPE_MASK       (1 << 8)
-#define SORT_TYPE_RESET      ~SORT_TYPE_MASK
-#define SORT_TYPE_DESC       (0 << 8)
-#define SORT_TYPE_ASC        (1 << 8)
+#else
+#define myprintf(args...) fprintf (args)
+#endif
 
-#define SORT_TYPE_GET(x)     (x & SORT_TYPE_MASK)
-#define SORT_TYPE_SET(x, v)  do{x &= SORT_TYPE_RESET; x |= v;}while(0)
-
-#define DETAIL_FILE_INFO (1 << 0)
-#define DETAIL_ELF_INFO  (1 << 1)
-#define DETAIL_SET(x,d) x |= d
-#define DETAIL_UNSET(x,d) x &= ~d
-
-typedef enum {
-  FORMAT_HUMAN = 0,
-  FORMAT_CSV,
-  FORMAT_CSV_SUMS
-} format_t;
-
-typedef struct
-{
-  gint      id;
-  address_t start;
-  guint     size;
-  gboolean  resident : 1;
-  gboolean  writable : 1;
-  gboolean  mapped   : 1;
-} page_t;
-
-typedef struct
-{
-  address_t  start;
-  address_t  end;
-  address_t  offset;
-  char       perms[5];
-  GList    * pages;
-  gchar    * file;
-  
-  gboolean   heap  : 1;
-  gboolean   stack : 1;
-  gboolean   vdso  : 1;
-  gboolean   anon  : 1;
-} vma_t;
-
-typedef struct
-{
-  pid_t        pid;
-  guint        vm;
-  guint        evm;
-  guint        heap;
-  guint        stack;
-  guint        vdso;
-  guint        mapped;
-  guint        emapped;
-  guint        sole;
-  guint        anon;
-  GHashTable * pg_count;
-  GHashTable * files;
-  GHashTable * map;
-  gchar      * cmd;
-  gboolean     load_symbols;
-} memory_t;
-
-typedef struct
-{
-  memory_t      m;
-  const gchar * name;
-  esym_t      * esym;
-}symbol_t;
-
-typedef struct
-{
-  memory_t      m;
-  const gchar * name;
-
-  GList * syms;
-}section_t;
-
-typedef struct
-{
-  memory_t   m;
-  elf_t    * elf;
-  GList    * vma;   /* of vma_t */
-  GList    * sects; /* of section_t */
-}file_t;
+static gint global_lines = 0;
 
 address_t
 page_start (address_t addr)
@@ -161,19 +85,6 @@ page_start (address_t addr)
   return addr & ~(getpagesize()-1);
 }
 
-typedef struct
-{
-  pid_t       * pids_arg;
-  int           pids_count;
-  char       ** pids_names;
-  guint         detail;
-  gboolean      all_pids;
-  guint         sort;
-  guint         threshold;
-  format_t      format;
-  const char  * log;
-} args_t;
-
 static void
 vma_free (gpointer p)
 {
@@ -212,7 +123,7 @@ memory_destroy (memory_t * m)
 
   if (m->files)
     g_hash_table_destroy (m->files);
-
+  
   g_free (m->cmd);
 }
 
@@ -226,6 +137,17 @@ memory_free (memory_t * m)
   g_free (m);
 }
 
+static void
+symbol_free (symbol_t * s)
+{
+  if (!s)
+    return;
+
+  memory_destroy (&(s->m));
+
+  g_free (s);
+}
+
 static void
 section_free (section_t * s)
 {
@@ -234,6 +156,20 @@ section_free (section_t * s)
 
   memory_destroy (&(s->m));
 
+  if (s->syms)
+    {
+      GList * l = s->syms;
+      while (l)
+        {
+          symbol_t * sym = l->data;
+          symbol_free (sym);
+      
+          l = l->next;
+        }
+
+      g_list_free (s->syms);
+    }
+  
   g_free (s);
 }
 
@@ -494,16 +430,16 @@ process_files (gpointer key, gpointer value, gpointer data)
 }
 
 static guint
-print_symbol_info (FILE * of, symbol_t * s, format_t format)
+print_symbol_info (FILE * of, symbol_t * s, data_t * d)
 {
   if (!s || !s->m.vm)
     return 0;
 
   guint lines = 0;
   
-  if (format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN)
     {
-      fprintf (of,"      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
+      myprintf (of, "      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
                s->name,
                SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
                SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
@@ -511,9 +447,9 @@ print_symbol_info (FILE * of, symbol_t * s, format_t format)
       
       lines++;
     }
-  else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+  else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      fprintf (of,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+      myprintf (of, ",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
                s->name, 
                s->m.vm,
                s->m.mapped,
@@ -527,7 +463,7 @@ print_symbol_info (FILE * of, symbol_t * s, format_t format)
 
 /* TODO refactor these print fuctions to share code */
 static guint
-print_section_info (FILE * of, section_t * s, format_t format)
+print_section_info (FILE * of, section_t * s, data_t * d)
 {
   if (!s)
     return 0;
@@ -536,9 +472,9 @@ print_section_info (FILE * of, section_t * s, format_t format)
 
   if (s->m.vm)
     {
-      if (format == FORMAT_HUMAN)
+      if (d->args.format == FORMAT_HUMAN)
         {
-          fprintf (of,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
+          myprintf (of,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
                    s->name,
                    SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
                    SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
@@ -546,9 +482,10 @@ print_section_info (FILE * of, section_t * s, format_t format)
       
           lines++;
         }
-      else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+      else if (d->args.format == FORMAT_CSV ||
+               d->args.format == FORMAT_CSV_SUMS)
         {
-          fprintf (of,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+          myprintf (of,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
                    s->name, 
                    s->m.vm,
                    s->m.mapped,
@@ -559,14 +496,15 @@ print_section_info (FILE * of, section_t * s, format_t format)
     }
   else if (s->syms)
     {
-      if (format == FORMAT_HUMAN)
+      if (d->args.format == FORMAT_HUMAN)
         {
-          fprintf (of,"   %-29s:\n", s->name);
+          myprintf (of,"   %-29s:\n", s->name);
           lines++;
         }
-      else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+      else if (d->args.format == FORMAT_CSV ||
+               d->args.format == FORMAT_CSV_SUMS)
         {
-          fprintf (of,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name);
+          myprintf (of,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name);
           lines++;
         }
     }
@@ -575,7 +513,7 @@ print_section_info (FILE * of, section_t * s, format_t format)
   while (l)
     {
       symbol_t * sbl = l->data;
-      lines += print_symbol_info (of, sbl, format);
+      lines += print_symbol_info (of, sbl, d);
       l = l->next;
     }
   
@@ -583,7 +521,7 @@ print_section_info (FILE * of, section_t * s, format_t format)
 }
 
 static guint
-print_file_info (FILE * of, file_t * f, format_t format, guint detail)
+print_file_info (FILE * of, file_t * f, data_t * d)
 {
   if (!f)
     return 0;
@@ -591,38 +529,41 @@ print_file_info (FILE * of, file_t * f, format_t format, guint detail)
   guint lines = 0;
   gchar * cmd = g_path_get_basename (f->m.cmd);
   
-  if (format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN)
     {
-      if (detail & DETAIL_ELF_INFO)
+      if (d->args.detail & DETAIL_ELF_INFO)
         {
-          fprintf (of,"  ----------------------------------"
+          myprintf (of,"  ----------------------------------"
                    "---------------------------------------\n");
           lines++;
         }
 
-      fprintf (of,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
+      myprintf (of,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
                cmd,
                SCALE_V(f->m.vm),     SCALE_U(f->m.vm),
                SCALE_V(f->m.mapped), SCALE_U(f->m.mapped),
                SCALE_V(f->m.sole),   SCALE_U(f->m.sole));
+      global_lines--;
+      
 #if 0
       elf_print_dbg_info (f->elf);
 #endif
-      if (detail & DETAIL_ELF_INFO)
+      if (d->args.detail & DETAIL_ELF_INFO)
         {
           lines += 2;
-          fprintf (of,"\n  ----------------------------------"
+          myprintf (of,"\n  ----------------------------------"
                    "---------------------------------------\n");
+          global_lines++;
         }
       else
         {
           lines++;
-          fprintf (of,"\n");
+          myprintf (of,"\n");
         }
     }
-  else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+  else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      fprintf (of,",,,,,,,,\"%s\",%u,%u,%u\n",
+      myprintf (of,",,,,,,,,\"%s\",%u,%u,%u\n",
                f->m.cmd, 
                f->m.vm,
                f->m.mapped,
@@ -633,13 +574,13 @@ print_file_info (FILE * of, file_t * f, format_t format, guint detail)
 
   g_free (cmd);
 
-  if (detail & DETAIL_ELF_INFO)
+  if (d->args.detail & DETAIL_ELF_INFO)
     {
       GList * l = f->sects;
       while (l)
         {
           section_t * s= (section_t*)l->data;
-          lines += print_section_info (of, s, format);
+          lines += print_section_info (of, s, d);
           l = l->next;
         }
     }
@@ -880,31 +821,27 @@ load_pid_data (pid_t pid,
           /* see if this process is among named processes given to us on the
            * cmd line
            */
-          if (args->pids_names && args->pids_names[0])
+          if (mem_data->cmd && *mem_data->cmd &&
+              args->pids_names && args->pids_names[0])
             {
               /* Unfortunately, g_path_basename() does not work if the
                * cmd line has any parameters, so we extract the process name
                * ourselves.
-               * 
-               * -- this assumes no spaces in file names
                */
-              gchar * cmd2 = g_strdup (mem_data->cmd);
-              gchar * c;
-              gchar * name;
-
-              if ((c = strrchr (cmd2, ' ')))
-                *c = 0;
-
-              if ((c = strrchr (cmd2, '/')))
-                name = c+1;
-              else
-                name = cmd2;
+              gchar * c = mem_data->cmd;
+              gchar * name = mem_data->cmd;
 
+              while ((c = strchr (c, '/')))
+                {
+                  c++;
+                  name = c;
+                }
+      
               char ** p = args->pids_names;
 
               while (*p)
                 {
-                  if (!strcmp (name, *p))
+                  if (!strncmp (name, *p, strlen (*p)))
                     {
                       /* add space for the new pid
                        * 2 == 1 for new pid 1 for NULL terminator
@@ -919,8 +856,6 @@ load_pid_data (pid_t pid,
                   
                   p++;
                 }
-
-              g_free (cmd2);
             }
           
         }
@@ -1091,6 +1026,7 @@ sort_pid_list (gconstpointer a, gconstpointer b, gpointer data)
       case SORT_BY_HEAP:   r = m2->heap - m1->heap;     break;
       case SORT_BY_STACK:  r = m2->stack - m1->stack;   break;
       case SORT_BY_MAPPED: r = m2->mapped - m1->mapped; break;
+      case SORT_BY_EMAPPED: r = m2->emapped - m1->emapped; break;
 
       default:
       case SORT_BY_NONE: r = 0;                   break;
@@ -1106,56 +1042,57 @@ sort_pid_list (gconstpointer a, gconstpointer b, gpointer data)
  * file_info indicates if per-file information is to be printed as well
  */
 static guint
-print_results (FILE * of, memory_t * mem_data, guint detail, guint threshold,
-               format_t format, int sort)
+print_results (FILE * of, memory_t * mem_data, data_t * d)
 {
   guint lines = 0;
 
   /* do not print if the vm is below the threshold, and or 0 */
-  threshold = threshold ? threshold : 1;
+  guint threshold = d->args.threshold ? d->args.threshold : 1;
   
   if (threshold && mem_data->vm < threshold)
     return lines;
 
-  if (format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN)
     {
-      fprintf (of,"Process %d", mem_data->pid);
-
+      myprintf (of,"Process %d", mem_data->pid);
+      global_lines--;
+      
       if (mem_data->cmd)
-        fprintf (of," [%s]\n", mem_data->cmd);
+        myprintf (of," [%s]\n", mem_data->cmd);
       else
-        fprintf (of,"\n");
+        myprintf (of,"\n");
   
-      fprintf (of," Virtual memory  : %9u %s\n",
+      myprintf (of," Virtual memory  : %9u %s\n",
                SCALE_V(mem_data->vm), SCALE_U(mem_data->vm));
-      fprintf (of," Effective VM    : %9u %s\n",
+      myprintf (of," Effective VM    : %9u %s\n",
                SCALE_V(mem_data->evm), SCALE_U(mem_data->evm));
-      fprintf (of," Heap            : %9u %s\n",
+      myprintf (of," Heap            : %9u %s\n",
                SCALE_V(mem_data->heap), SCALE_U(mem_data->heap));
-      fprintf (of," Stack           : %9u %s\n",
+      myprintf (of," Stack           : %9u %s\n",
                SCALE_V(mem_data->stack), SCALE_U(mem_data->stack));
-      fprintf (of," vdso            : %9u %s\n",
+      myprintf (of," vdso            : %9u %s\n",
                SCALE_V(mem_data->vdso), SCALE_U(mem_data->vdso));
-      fprintf (of," Mapped          : %9u %s\n",
+      myprintf (of," Mapped          : %9u %s\n",
                SCALE_V(mem_data->mapped), SCALE_U(mem_data->mapped));
-      fprintf (of," Effective mapped: %9u %s\n",
+      myprintf (of," Effective mapped: %9u %s\n",
                SCALE_V(mem_data->emapped), SCALE_U(mem_data->emapped));
-      fprintf (of," Sole use        : %9u %s\n",
+      myprintf (of," Sole use        : %9u %s\n",
                SCALE_V(mem_data->sole), SCALE_U(mem_data->sole));
 
       lines += 9;
       
-      if (detail & DETAIL_FILE_INFO)
+      if (d->args.detail & DETAIL_FILE_INFO)
         {
           lines += 2;
-          fprintf (of,"\n Per file memory use\n");
+          global_lines++;
+          myprintf (of,"\n Per file memory use\n");
         }
     }
-  else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+  else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
       lines++;
       
-      fprintf (of,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
+      myprintf (of,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
                mem_data->pid,
                mem_data->cmd ? mem_data->cmd : "???",
                mem_data->vm,
@@ -1168,17 +1105,18 @@ print_results (FILE * of, memory_t * mem_data, guint detail, guint threshold,
                mem_data->sole);
     }
 
-  if (detail & DETAIL_FILE_INFO)
+  if (d->args.detail & DETAIL_FILE_INFO)
     {
       GList * l = NULL;
       g_hash_table_foreach (mem_data->files, hash_to_list, &l);
 
       if (l)
         {
-          l = g_list_sort_with_data (l, sort_pid_list, GINT_TO_POINTER (sort));
+          l = g_list_sort_with_data (l, sort_pid_list,
+                                     GINT_TO_POINTER (d->args.sort));
           while (l)
             {
-              lines += print_file_info (of, (file_t*)l->data, format, detail);
+              lines += print_file_info (of, (file_t*)l->data, d);
               l = l->next;
             }
 
@@ -1186,10 +1124,16 @@ print_results (FILE * of, memory_t * mem_data, guint detail, guint threshold,
         }
     }
 
-  if (format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN)
     {
-      fprintf (of,"\n========================================"
-               "=======================================\n\n");
+      if (!d->args.interactive)
+        {
+          myprintf (of,"\n========================================"
+                       "=======================================\n\n");
+          global_lines += 2;
+        }
+      else
+        myprintf (of, "\n");
     }
 
   return lines;
@@ -1230,7 +1174,8 @@ usage (const char * option)
   fprintf (f, "  --sort-by=what        : vm|sole|heap|stack|mapped|none (default: none)\n");
   fprintf (f, "  --sort-type=which     : desc|asc (default: desc)\n");
   fprintf (f, "  --format=which        : human|csv|csv+sums (default: human)\n");
-  fprintf (f, "  --log=file            : print results to file, not stdout\n");
+  fprintf (f, "  --log=file            : Print results to file, not stdout\n");
+  fprintf (f, "  --interactive         : Run in interactive mode\n");
   fprintf (f, "  --version             : Program version\n");
   fprintf (f, "  --module-version      : Version of required kernel module\n");
   fprintf (f, "  --help\n\n");
@@ -1243,7 +1188,7 @@ usage (const char * option)
   fprintf (f, "processes specified explicitely symbol information is printed where available.\n\n");
 }
 
-static gboolean
+static gint
 parse_args (int argc, char ** argv, args_t * args)
 {
   gint i, j;
@@ -1251,7 +1196,7 @@ parse_args (int argc, char ** argv, args_t * args)
   if (argc < 2)
     {
       usage (NULL);
-      return FALSE;
+      return 0;
     }
 
   for (i = 1; i < argc; ++i)
@@ -1289,6 +1234,8 @@ parse_args (int argc, char ** argv, args_t * args)
             SORT_BY_SET (args->sort, SORT_BY_STACK);
           else if (!strcmp (argv[i] + 10, "mapped"))
             SORT_BY_SET (args->sort, SORT_BY_MAPPED);
+          else if (!strcmp (argv[i] + 10, "emapped"))
+            SORT_BY_SET (args->sort, SORT_BY_EMAPPED);
           else
             SORT_BY_SET (args->sort, SORT_BY_NONE);
 
@@ -1326,9 +1273,20 @@ parse_args (int argc, char ** argv, args_t * args)
       if (!strncmp (argv[i], "--log",5))
         {
           args->log = argv[i] + 6;
-
           continue;
         }
+
+      if (!strcmp (argv[i], "--interactive"))
+        {
+#ifdef INTERACTIVE
+          args->interactive = TRUE;
+          continue;
+#else
+          g_print ("--interactive not supported "
+                   "(compiled with --without-readline)\n");
+          return EXIT_INVALID_ARGS;
+#endif
+        }
       
       if (!strcmp (argv[i], "--help"))
         {
@@ -1351,9 +1309,16 @@ parse_args (int argc, char ** argv, args_t * args)
     args_error:
       /* unknown option */
       usage (argv[i]);
-      return FALSE;
+      return EXIT_INVALID_ARGS;
     }
-  
+
+  if (args->interactive &&
+      (args->all_pids || args->log))
+    {
+      g_print ("Cannot use --all-pids or --log with --interactive.\n");
+      return EXIT_INVALID_ARGS;
+    }
+
   args->pids_arg   = g_malloc0((argc - i + 1) * sizeof(pid_t));
   args->pids_names = g_malloc0((argc - i + 1) * sizeof(const char*));
   args->pids_count = 0;
@@ -1373,7 +1338,14 @@ parse_args (int argc, char ** argv, args_t * args)
         }
     }
 
-  return TRUE;
+  if (!j && !args->all_pids && !args->pids_count && !args->interactive)
+    {
+      g_warning ("Either use --all-pids, or specify at least one pid");
+      return EXIT_INVALID_ARGS;
+    }
+  
+
+  return 0;
 }
 
 static gboolean
@@ -1390,55 +1362,26 @@ is_pid_present (pid_t * pids, pid_t pid)
   return FALSE;
 }
 
-int
-main (int argc, char ** argv)
+void
+main_load (data_t * d)
 {
-  GHashTable * pg_count = NULL;
-  GList      * pids_sys = NULL;
-  pid_t      * pid;
-  GList      * mem_list = NULL;
   GList      * l;
-  gint         retval = 0;
-  guint        lines = 0;
-  FILE *       of = stdout;
-  args_t       args;
-  time_t       now;
-  struct tm    t;
-  char         date_stamp[30];
-  
-  args.pids_arg   = NULL;
-  args.pids_names = NULL;
-  args.detail     = (DETAIL_FILE_INFO | DETAIL_ELF_INFO);
-  args.all_pids   = FALSE;
-  args.sort       = 0;
-  args.threshold  = 0;
-  args.format     = FORMAT_HUMAN;
-  args.log        = NULL;
 
-  if (! parse_args (argc, argv, &args))
+  if (d->args.interactive)
     {
-      retval = -1;
-      goto finish;
-    }
-
-  if (args.log && (of = fopen (args.log, "w")) == NULL)
-    {
-      g_warning ("Could not open putput file [%s]", args.log);
-      retval = EXIT_LOGFILE_ERROR;
-      goto finish;
+      g_print ("  Creating system snapshot ...");
     }
   
-  if (! (pids_sys = load_pids ()))
+  if (!(d->pids_sys = load_pids ()))
     {
       g_warning ("There was a problem loading system data");
-      retval = EXIT_SYSDATA_ERROR;
-      goto finish;
+      exit (EXIT_SYSDATA_ERROR);
     }
 
-  pg_count = g_hash_table_new (NULL, NULL);
-
+  d->pg_count = g_hash_table_new (NULL, NULL);
+  
   /* load data for each accessible pid */
-  l = pids_sys;
+  l = d->pids_sys;
   while (l)
     {
       memory_t * m;
@@ -1449,10 +1392,10 @@ main (int argc, char ** argv)
       if (!h)
         continue;
       
-      if ((m = load_pid_data (p, h, pg_count, &args)))
+      if ((m = load_pid_data (p, h, d->pg_count, &(d->args))))
         {
-          m->load_symbols = is_pid_present (args.pids_arg, p);
-          mem_list = g_list_prepend (mem_list, m);
+          m->load_symbols = is_pid_present (d->args.pids_arg, p);
+          d->mem_list = g_list_prepend (d->mem_list, m);
         }
       
       l = l->next;
@@ -1462,7 +1405,7 @@ main (int argc, char ** argv)
    * and how much between files.
    */
   
-  l = mem_list;
+  l = d->mem_list;
   while (l)
     {
       memory_t * m = (memory_t *) l->data;
@@ -1472,101 +1415,144 @@ main (int argc, char ** argv)
 
       l = l->next;
     }
+}
 
+void
+main_sort (data_t * d)
+{
   /* sort results as appropriate */
-  if (SORT_BY_GET (args.sort) != SORT_BY_NONE)
+  if (SORT_BY_GET (d->args.sort) != SORT_BY_NONE)
     {
-      mem_list = g_list_sort_with_data (mem_list, sort_pid_list,
-                                        GINT_TO_POINTER (args.sort));
+      d->mem_list = g_list_sort_with_data (d->mem_list, sort_pid_list,
+                                           GINT_TO_POINTER (d->args.sort));
     }
+}
 
-  /* Print a time stamp into the log */
-  now = time (NULL);
-  localtime_r (&now, &t);
-  strftime (date_stamp, sizeof (date_stamp),
-            "%a %Y %m %d %H:%M:%S", &t);
+void
+main_print (data_t * d)
+{
+  guint        lines = 0;
+  GList      * l;
+  pid_t      * pid;
+  FILE       * of = d->of;
+
+  global_lines = 0;
+  
+  if (!d->args.interactive)
+    {
+      time_t       now;
+      struct tm    t;
+      char         date_stamp[30];
+  
+      /* Print a time stamp into the log */
+      now = time (NULL);
+      localtime_r (&now, &t);
+      strftime (date_stamp, sizeof (date_stamp),
+                "%a %Y %m %d %H:%M:%S", &t);
                       
-  fprintf (of,"--- Exmap log from %s ---\n\n", date_stamp);
+      fprintf (of,"--- Exmap log from %s ---\n\n", date_stamp);
+    }
   
   /* output header */
-  if (args.format == FORMAT_CSV || args.format == FORMAT_CSV_SUMS)
+  if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      fprintf (of,"pid,command,VM,effective VM,heap,stack,vdso,mapped,sole");
-
-      if (args.detail & DETAIL_FILE_INFO)
+      myprintf (of,
+               "pid,command,VM,effective VM,heap,stack,vdso,mapped,"
+               "emapped,sole");
+      global_lines--;
+      
+      if (d->args.detail & DETAIL_FILE_INFO)
         {
-          fprintf (of,",file,file-VM,file-mapped,file-solefile");
+          myprintf (of,",file,file-VM,file-mapped,file-solefile");
 
-          if (args.detail & DETAIL_ELF_INFO)
-            fprintf (of,",section,section-VM,section-mapped,section-sole\n");
+          if (d->args.detail & DETAIL_ELF_INFO)
+            myprintf (of,
+                     ",section,section-VM,section-mapped,section-sole\n");
           else
-            fprintf (of,"\n");
+            myprintf (of,"\n");
+
+          global_lines--;
         }
       else
-        fprintf (of,"\n");
+        myprintf (of,"\n");
     }
 
   /* output info for individula pids */
-  if (args.all_pids)
+  if (d->args.all_pids)
     {
-      l = mem_list;
+      l = d->mem_list;
       while (l)
         {
-          lines += print_results (of, (memory_t*)l->data,
-                                  args.detail, args.threshold, args.format, args.sort);
+          lines += print_results (d->of, (memory_t*)l->data, d);
           l = l->next;
         }
     }
   else
     {
-      pid = args.pids_arg;
-      while (*pid)
+      pid = d->args.pids_arg;
+      if (pid && *pid)
         {
-          l = g_list_find_custom (mem_list, GINT_TO_POINTER (*pid), find_pid);
+          while (*pid)
+            {
+              l = g_list_find_custom (d->mem_list,
+                                      GINT_TO_POINTER (*pid), find_pid);
 
-          if (l)
-            lines += print_results (of, (memory_t*)l->data,
-                                    args.detail, args.threshold,
-                                    args.format, args.sort);
-          pid++;
+              if (l)
+                lines += print_results (d->of, (memory_t*)l->data, d);
+              pid++;
+            }
+        }
+      else
+        {
+          fprintf (d->of, "  No process selected.\n");
         }
     }
 
-  if (args.format == FORMAT_CSV_SUMS)
+  if (!d->args.interactive)
     {
-      /* output sum line -- keep in sync with the header line */
-      int i;
-      char c = 'A' + 2;
+      if (d->args.format == FORMAT_CSV_SUMS)
+        {
+          /* output sum line -- keep in sync with the header line */
+          int i;
+          char c = 'A' + 2;
 
-      fprintf (of,"Totals,\"(F9 to recalculate)\""); /* in place of pid, cmd */
-      
-      for (i = 0; i < CSV_COLUMNS - 2; ++i, ++c)
-        fprintf (of,",\"=sum(%c2:%c%d)\"", c, c, lines+1);
+          fprintf (d->of,"Totals,\"(F9 to recalculate)\"");
+          
+          for (i = 0; i < CSV_COLUMNS - 2; ++i, ++c)
+            fprintf (d->of,",\"=sum(%c2:%c%d)\"", c, c, lines+1);
 
-      fprintf (of,"\n");
+          fprintf (d->of,"\n");
 
-      lines++;
+          lines++;
 
-      if (lines > G_MAXUINT16)
-        fprintf (stderr,
-                 "\n"
-                 "   WARNING: Generated CSV file contain %u lines -- spreadsheet\n"
-                 "            application may not be able to load this file in its\n"
-                 "            entirety.\n\n",
-                 lines);
+          if (lines > G_MAXUINT16)
+            fprintf (stderr,
+                     "\n"
+                     "   WARNING: Generated CSV file contain %u lines -- spreadsheet\n"
+                     "            application may not be able to load this file in its\n"
+                     "            entirety.\n\n",
+                     lines);
+        }
+
+      /* put a mark in, so we know that the process finished normally */
+      fprintf (d->of,"--- Exmap log ends here ---\n");
     }
+}
 
-  /* put a mark in, se we know that the process finished normally */
-  fprintf (of,"--- Exmap log ends here ---\n");
+void
+main_reset (data_t * d)
+{
+  GList * l;
   
- finish:
-  /* clean up */
-  if (pg_count)
-    g_hash_table_destroy (pg_count);
-  
-  if (mem_list)
+  if (d->pg_count)
     {
-      l = mem_list;
+      g_hash_table_destroy (d->pg_count);
+      d->pg_count = NULL;
+    }
+  
+  if (d->mem_list)
+    {
+      l = d->mem_list;
       while (l)
         {
           memory_t * m = (memory_t *) l->data;
@@ -1574,16 +1560,74 @@ main (int argc, char ** argv)
           l = l->next;
         }
 
-      g_list_free (mem_list);
+      g_list_free (d->mem_list);
+      d->mem_list = NULL;
     }
 
-  if (pids_sys)
-    g_list_free (pids_sys);
+  if (d->pids_sys)
+    {
+      g_list_free (d->pids_sys);
+      d->pids_sys = NULL;
+    }
+}
 
-  if (args.pids_arg)
-    g_free (args.pids_arg);
+int
+main (int argc, char ** argv)
+{
+  gint         retval = 0;
+  data_t       data;
+  
+  memset (&data, 0, sizeof (data));
+  
+  data.of          = stdout;
+  data.args.detail = (DETAIL_FILE_INFO | DETAIL_ELF_INFO);
+  data.args.format = FORMAT_HUMAN;
 
-  fclose (of);
+  if ((retval = parse_args (argc, argv, &(data.args))))
+    {
+      goto finish;
+    }
+
+  if (data.args.log && (data.of = fopen (data.args.log, "w")) == NULL)
+    {
+      g_warning ("Could not open putput file [%s]", data.args.log);
+      retval = EXIT_LOGFILE_ERROR;
+      goto finish;
+    }
+
+#ifdef INTERACTIVE
+  if (data.args.interactive)
+    {
+      interactive_init (&data);
+    }
+#endif
+  
+  main_load (&data);
+  main_sort (&data);
+
+#ifdef INTERACTIVE
+  if (data.args.interactive)
+    {
+      g_print (" %d processes running.\n"
+               "  Enter a command or '?' for help.\n",
+               g_list_length (data.mem_list));
+      
+      interactive_loop (&data);
+    }
+  else
+#endif
+    main_print (&data);
+  
+ finish:
+  main_reset (&data);
+
+  if (data.args.pids_arg)
+    g_free (data.args.pids_arg);
+
+  if (data.args.pids_names)
+    g_free (data.args.pids_names);
+  
+  fclose (data.of);
   
   return retval;
 }
